from __future__ import annotations

from typing import List, Optional

from assemblyline import odm
from assemblyline.odm.models.workflow import PRIORITIES, STATUSES

ES_SUBMITTED = "submitted"
ES_SKIPPED = "skipped"
ES_INCOMPLETE = "incomplete"
ES_COMPLETED = "completed"
EXTENDED_SCAN_VALUES = {ES_SUBMITTED, ES_SKIPPED, ES_INCOMPLETE, ES_COMPLETED}
EXTENDED_SCAN_PRIORITY = [ES_COMPLETED, ES_INCOMPLETE, ES_SKIPPED, ES_SUBMITTED]


def merge_extended_scan(a: str, b: str) -> str:
    # Select the prefered value
    for value in EXTENDED_SCAN_PRIORITY:
        if a == value or b == value:
            return value
    raise ValueError(f"Invalid program state. scan state {a} {b}")


DetailedItemSubtype = odm.Enum(
    [
        'CFG',  # Configuration blocks
        'EXP',  # Exploits
        'IMP',  # Implants
        'OB',   # Obfuscation methods
        'TA',   # Threat actors
    ],
    description="Specifies the item's subtype (e.g., EXP, CFG, OB, IMP, TA)."
)

@odm.model(index=True, store=False, description="""Represents a granular element within the detailed analysis results, providing specific insights into the analysis findings.
""")
class DetailedItem(odm.Model):
    type = odm.Keyword(description="Defines the specific attribute or aspect of the analysis that this detailed item pertains to.")
    value = odm.Keyword(description="The specific value or identifier for the detail item.")
    verdict = odm.Enum(['safe', 'info', 'suspicious', 'highly suspicious', 'malicious'], description="Security assessment of the detailed item.")
    subtype = odm.Optional(DetailedItemSubtype, description="Specifies the item's subtype (e.g., CFG, EXP, IMP, OB, TA).")
    
    def __hash__(self) -> int:
        return hash(tuple(sorted(self.as_primitives().items())))

    def __lt__(self, other: DetailedItem) -> bool:
        if self.type != other.type:
            return self.type < other.type
        if self.value != other.value:
            return self.value < other.value
        if self.verdict != other.verdict:
            return self.verdict < other.verdict
        return self.subtype < other.subtype


@odm.model(index=True, store=False, description="""Stores information about screenshots taken during the analysis of the file. Each screenshot has a name, description, and the hashes of the image and its thumbnail, offering a visual reference that can aid in manual review processes.
""")
class Screenshot(odm.Model):
    name = odm.Keyword(description="The name or title of the screenshot.")
    description = odm.Keyword(description="A brief description of the screenshot's content.")
    img = odm.SHA256(description="The SHA256 hash of the full-size screenshot image.")
    thumb = odm.SHA256(description="The SHA256 hash of the thumbnail version of the screenshot.")

    def __hash__(self) -> int:
        return hash(tuple(sorted(self.as_primitives().items())))


@odm.model(index=True, store=False, description="""Provides a comprehensive breakdown of specific attributes and their associated analysis results.
""")
class DetailedResults(odm.Model):
    attack_pattern = odm.List(odm.Compound(DetailedItem), default=[], description="MITRE ATT&CK® framework patterns identified in the analysis.")
    attack_category = odm.List(odm.Compound(DetailedItem), default=[], description="MITRE ATT&CK® framework categories associated with the alert.")
    attrib = odm.List(odm.Compound(DetailedItem), default=[], description="Attribution information that provides context by suggesting associations with known malware families, suspected threat actors, or ongoing campaigns.")
    av = odm.List(odm.Compound(DetailedItem), default=[], description="Information on antivirus signature matches.")
    behavior = odm.List(odm.Compound(DetailedItem), default=[], description="Descriptions of the behaviors exhibited by the analyzed file or artifact that led to the alert.")
    domain = odm.List(odm.Compound(DetailedItem), default=[], description="Domain information related to the alert.")
    heuristic = odm.List(odm.Compound(DetailedItem), default=[], description="Heuristic information that triggered the alert.")
    ip = odm.List(odm.Compound(DetailedItem), default=[], description="IP address information related to the alert.")
    uri = odm.List(odm.Compound(DetailedItem), default=[], description="URI information related to the alert.")
    yara = odm.List(odm.Compound(DetailedItem), default=[], description="Information on YARA rule matches that contributed to the alert.")

    def update(self, other: DetailedResults) -> None:
        for field in self.fields().keys():
            setattr(self, field, list(set(getattr(self, field) + getattr(other, field))))


@odm.model(index=True, store=False, description="""Contains the aggregated results of the analysis performed by Assemblyline. It includes information such as attribution, behaviors observed, domains and IPs related to the threat, and the overall score indicating the severity of the findings.
""")
class ALResults(odm.Model):
    attrib = odm.List(odm.Keyword(), default=[], store=True, copyto="__text__", description="A list of attribution tags that provide context by suggesting associations with known malware families, suspected threat actors, or ongoing campaigns.")
    av = odm.List(odm.Keyword(), default=[], store=True, copyto="__text__", description="List of antivirus signatures that matched the file associated with the alert.")
    behavior = odm.List(odm.Keyword(), default=[], copyto="__text__", description="Descriptions of behaviors exhibited by the analyzed file or artifact that led to the alert.")
    detailed = odm.Compound(DetailedResults, description="Provides a more detailed breakdown of the analysis results.")
    domain = odm.List(odm.Domain(), default=[], copyto="__text__", description="Aggregate list of domains related to the alert, derived from both static and dynamic analysis.")
    domain_dynamic = odm.List(odm.Domain(), default=[], description="List of domains observed during dynamic analysis of the artifact.")
    domain_static = odm.List(odm.Domain(), default=[], description="List of domains extracted from static analysis of the artifact.")
    ip = odm.List(odm.IP(), default=[], copyto="__text__", description="Aggregate list of IP addresses related to the alert, derived from both static and dynamic analysis.")
    ip_dynamic = odm.List(odm.IP(), default=[], description="List of IP addresses observed during dynamic analysis of the artifact.")
    ip_static = odm.List(odm.IP(), default=[], description="List of IP addresses extracted from static analysis of the artifact.")
    request_end_time = odm.Date(index=False, description="The timestamp indicating when the processing of the submission completed.")
    score = odm.Integer(store=True, description="The highest score assigned to any part of the submission based on the analysis results.")
    uri = odm.List(odm.URI(), default=[], copyto="__text__", description="Aggregate list of URIs related to the alert, derived from both static and dynamic analysis.")
    uri_dynamic = odm.List(odm.URI(), default=[], description="List of URIs observed during dynamic analysis of the artifact.")
    uri_static = odm.List(odm.URI(), default=[], description="List of URIs extracted from static analysis of the artifact.")
    yara = odm.List(odm.Keyword(), default=[], copyto="__text__", description="List of YARA rule matches that contributed to the alert.")

    def update(self, other: ALResults) -> None:
        # Handle the fields that require special treatment
        fields = list(self.fields().keys())
        fields.remove('detailed')
        self.detailed.update(other.detailed)
        fields.remove('request_end_time')
        self.request_end_time = max(self.request_end_time, other.request_end_time)
        fields.remove('score')
        self.score = max(self.score, other.score)

        # All of the rest can simply be merged
        for field in fields:
            setattr(self, field, list(set(getattr(self, field) + getattr(other, field))))


@odm.model(index=True, store=True, description="""Captures comprehensive metadata and unique identifiers for the original file submitted for analysis, which is central to the generation of the alert.
""")
class File(odm.Model):
    md5 = odm.MD5(copyto="__text__", description="The MD5 hash of the file.")
    name = odm.Keyword(copyto="__text__", description="The original name of the file as submitted.")
    sha1 = odm.SHA1(copyto="__text__", description="The SHA1 hash of the file.")
    sha256 = odm.SHA256(copyto="__text__", description="The SHA256 hash of the file.")
    size = odm.long(store=False, description="The size of the file in bytes.")
    type = odm.Keyword(copyto="__text__", description="	The file type as identified by Assemblyline's analysis.")
    screenshots = odm.List(odm.Compound(Screenshot), default=[], description="Screenshots taken of the file during analysis, if applicable.")

    def update(self, other: File) -> None:
        if self.sha256 != other.sha256:
            raise ValueError(f"Only alerts on the same file may be merged {self.sha256} != {other.sha256}")
        self.screenshots = list(set(self.screenshots + other.screenshots))


@odm.model(index=True, store=False, description="""The Verdict submodel captures the conclusions drawn by users regarding the nature of a submission. It lists user identifiers for those who have deemed the submission as either malicious or non-malicious, representing a collective assessment of the threat.
""")
class Verdict(odm.Model):
    malicious = odm.List(odm.Keyword(), default=[], description="User IDs of those who have marked the alert as malicious.")
    non_malicious = odm.List(odm.Keyword(), default=[],
                             description="User IDs of those who have marked the alert as non-malicious.")

    def update(self, other: Verdict) -> None:
        self.malicious = list(set(self.malicious + other.malicious))
        self.non_malicious = list(set(self.non_malicious + other.non_malicious))


@odm.model(index=True, store=False, description="""Summarizes the heuristics that were triggered during the analysis. These heuristics are part of the detection logic used by Assemblyline to identify suspicious or malicious behavior in the analyzed file.
""")
class Heuristic(odm.Model):
    name = odm.List(odm.Keyword(), default=[], description="Names of the heuristics that have been matched in the analysis.")

    def update(self, other: Heuristic) -> None:
        self.name = list(set(self.name + other.name))


@odm.model(index=True, store=False, description="""The Attack submodel is a component of the Alert model that records information aligned with the MITRE ATT&CK framework. It lists the ATT&CK patterns and categories that have been identified in the analysis, helping to map the threat to known adversary tactics and techniques.
""")
class Attack(odm.Model):
    pattern = odm.List(odm.Keyword(), default=[], description="List of MITRE ATT&CK® framework patterns that are relevant to the alert.")
    category = odm.List(odm.Keyword(), default=[], description="List of MITRE ATT&CK® framework categories that are relevant to the alert.")

    def update(self, other: Attack) -> None:
        self.pattern = list(set(self.pattern + other.pattern))
        self.category = list(set(self.category + other.category))


@odm.model(index=True, store=False, description="Describes an event or action that has occurred during the lifecycle of the alert, capturing changes in status, priority, or labels.")
class Event(odm.Model):
    entity_type: str = odm.Enum(values=['user', 'workflow'], description="The type of entity associated with the event.")
    entity_id: str = odm.Keyword(description="The unique identifier of the entity associated with the event.")
    entity_name: str = odm.Keyword(description="The name of the entity associated with the event.")
    ts: str = odm.Date(default="NOW", description="The timestamp when the event occurred.")
    labels: List[str] = odm.sequence(odm.keyword(), default=[], description="Labels that were added to the alert during the event.")
    labels_removed: List[str] = odm.sequence(odm.keyword(), default=[], description="Labels that were removed from the alert during the event.")
    status: str = odm.Optional(odm.Enum(values=STATUSES), description="The status of the alert after the event took place.")
    priority: str = odm.Optional(odm.Enum(values=PRIORITIES), description="The priority level assigned to the alert during the event.")

    def __hash__(self) -> int:
        data = self.as_primitives()
        data['labels'] = tuple(sorted(self.labels))
        data['labels_removed'] = tuple(sorted(self.labels_removed))
        return hash(tuple(sorted(data.items())))


@odm.model(index=True, store=True, description="Describes the relationship between different submissions that are linked to the formation of the alert, highlighting parent-child connections.")
class Relationship(odm.Model):
    child: str = odm.UUID(description="The identifier of the child submission in the relationship.")
    parent: Optional[str] = odm.optional(odm.UUID(), description="The identifier of the parent submission, if applicable.")

    def __hash__(self) -> int:
        return hash(tuple(sorted(self.as_primitives().items())))


@odm.model(index=True, store=True, description="""The Alert object model, as defined in this documentation, specifies the structured representation of alert data within the Assemblyline application's Alert index. Each field delineated in the schema is an attribute of the Alert document, characterized by its data type, semantic definition, mandatory status, and default instantiation value.

Comprehension of this schema is pivotal for the construction of targeted Lucene search queries, which are instrumental in the interrogation and retrieval of alert-specific data from Assemblyline's analytical output. The schema's fields provide the analytical lexicon necessary to query and dissect alert data, facilitating the isolation of alerts based on defined parameters such as threat identifiers, heuristic evaluations, and temporal metadata.

This schema serves as a technical blueprint for cybersecurity professionals to navigate Assemblyline's alerting system, enabling refined query strategies and data extraction methodologies that align with operational cybersecurity imperatives and threat intelligence workflows.
""")
class Alert(odm.Model):
    alert_id = odm.Keyword(copyto="__text__", description="Unique identifier for the alert.")
    al = odm.compound(ALResults, description="Contains the results of the Assemblyline analysis for the alert.")
    archive_ts = odm.Optional(odm.Date(), description="Timestamp indicating when the alert was archived in the system.")
    attack = odm.Compound(Attack, description="Structured data representing MITRE ATT&CK information associated with the alert.")
    classification = odm.Classification(description="Security classification assigned to the alert based on its contents and context.")
    expiry_ts = odm.Optional(odm.Date(store=False), description="Timestamp indicating when the alert is scheduled to expire from the system.")
    extended_scan = odm.Enum(values=EXTENDED_SCAN_VALUES, description="Indicates the status of an extended scan, if applicable. Extended scans are additional analyses performed after the initial analysis.")
    file = odm.Compound(File, description="Information about the file associated with the alert.")
    filtered = odm.Boolean(default=False, description="Indicates whether portions of the submission's analysis results have been omitted due to the user's classification level not meeting the required threshold for viewing certain data.")
    heuristic = odm.Compound(Heuristic, description="Data regarding the heuristics that triggered the alert.")
    label = odm.List(odm.Keyword(), copyto="__text__", default=[], description="Labels assigned to the alert for categorization and filtering.")
    metadata: dict[str, str] = odm.FlatMapping(odm.MetadataValue(), default={}, store=False, copyto="__text__", description="Additional metadata provided with the file at the time of submission.")
    owner = odm.Optional(odm.Keyword(), description="Specifies the user or system component that has taken ownership of the alert. If no user has claimed the alert, it remains under system ownership with no specific user associated, indicated by a value of `None`.")
    priority = odm.Optional(odm.Enum(values=PRIORITIES), description="Indicates the importance level assigned to the alert.")
    reporting_ts = odm.Date(description="Timestamp when the alert was created.")
    submission_relations = odm.sequence(odm.compound(Relationship), description="Describes the hierarchical relationships between submissions that contributed to this alert.")
    sid = odm.UUID(description="Identifier for the submission associated with this alert.")
    status = odm.Optional(odm.Enum(values=STATUSES), description="Reflects the current state of the alert throughout its lifecycle. This status is subject to change as a result of user actions, automated processes, or the execution of workflows within Assemblyline. The status provides insight into the current phase of analysis or response.")
    ts = odm.Date(description="Timestamp of when the file submission occurred that led to the generation of this alert.")
    type = odm.Keyword(description="The type or category of the alert as specified at submission time by the user.")
    verdict = odm.Compound(Verdict, default={}, description="Consolidates user assessments of the submission's nature. It records the user identifiers of those who have evaluated the submission, categorizing it as either malicious or non-malicious.")
    events = odm.sequence(odm.compound(Event), default=[], description="An audit trail of events and actions taken on the alert.")
    workflows_completed = odm.Boolean(default=False, description="Flag indicating whether all configured workflows have been executed for this alert.")

    def update(self, other: Alert) -> None:
        """Update the current object given the content of a second alert."""

        # Make sure we are merging compatible alerts
        if self.alert_id != other.alert_id:
            raise ValueError(f"Only versions of the same alert may be merged id {self.alert_id} != {other.alert_id}")

        # Merge simple compounds using their own logic
        self.al.update(other.al)
        self.attack.update(other.attack)
        self.file.update(other.file)
        self.heuristic.update(other.heuristic)
        self.verdict.update(other.verdict)

        # Merge by reasonable simple operations on some fields
        self.classification = self.classification.max(other.classification)

        # If both are set to expire go with the furthest expiry,
        # otherwise if either is permanent make the product permanent
        if self.expiry_ts and other.expiry_ts:
            self.expiry_ts = max(self.expiry_ts, other.expiry_ts)
        else:
            self.expiry_ts = None

        # Prefer anything that isn't submitted
        self.extended_scan = merge_extended_scan(self.extended_scan, other.extended_scan)

        # If either is filtered, the content should be considered filtered
        self.filtered |= other.filtered
        # Keep all unique labels and metadata, where metadata doesn't match take one more or less at random
        self.label = list(set(self.label + other.label))
        self.metadata.update(other.metadata)

        # Prefer the current owner/priority/status, but take the other if the current isn't defined
        self.owner = self.owner or other.owner
        self.priority = self.priority or other.priority
        self.reporting_ts = max(self.reporting_ts, other.reporting_ts)
        self.ts = min(self.ts, other.ts)
        self.status = self.status or other.status

        # self.type is fine

        # Merge the submission list and update sid to a value that isn't a parent
        self.submission_relations = list(set(self.submission_relations + other.submission_relations))
        parents = set(relation.parent for relation in self.submission_relations if relation.parent)
        for relation in self.submission_relations:
            if relation.child not in parents and relation.parent:
                self.sid = relation.child

        # Merge the events then sort them by time
        self.events = list(sorted(set(self.events + other.events), key=lambda e: e.ts))

        # Always consider the updated alert a new one WRT workflows
        self.workflows_completed = False



from assemblyline import odm
from assemblyline.common import forge
from assemblyline.odm.models.statistics import Statistics

Classification = forge.get_classification()

DEPLOYED_STATUSES = ['DEPLOYED', 'NOISY', 'DISABLED']
DRAFT_STATUSES = ['STAGING', 'TESTING']
STALE_STATUSES = ['INVALID']

RULE_STATUSES = DEPLOYED_STATUSES + DRAFT_STATUSES + STALE_STATUSES


@odm.model(index=True, store=True)
class Signature(odm.Model):
    classification = odm.Classification(store=True, default=Classification.UNRESTRICTED, description="Security classification assigned to the signature based on its contents and context.")
    data = odm.Text(copyto="__text__", store=False)
    last_modified = odm.Date(default="NOW", description="Notes the last modification timestamp of the signature.")
    name = odm.Keyword(copyto="__text__", description="Name of the signature.")
    order = odm.Integer(default=1, store=False, deprecation="no longer used in v4")
    revision = odm.Keyword(default="1", description="")
    signature_id = odm.Optional(odm.Keyword(), description="ID associated with the signature.")
    source = odm.Keyword(description="Source or author of the signature.")
    state_change_date = odm.Optional(odm.Date(store=False), description="Date the signature's state was last changed.")
    state_change_user = odm.Optional(odm.Keyword(store=False), description="User who last changed the signature's state.")
    stats = odm.Compound(Statistics, default={}, description="Stats associated with count, average, min, max, and sum of various signature metrics.")
    status = odm.Enum(values=RULE_STATUSES, copyto="__text__", description="The current state of the signature (i.e. NOISY, DISABLED, DEPLOYED, etc.).")
    type = odm.Keyword(copyto="__text__", description="The service type that the signature is associated with.")




from assemblyline import odm

MSG_TYPES = {"RetrohuntHeartbeat"}
LOADER_CLASS = "assemblyline.odm.messages.retrohunt_heartbeat.RetrohuntMessage"


@odm.model(description="Heartbeat Model for retrohunt")
class Heartbeat(odm.Model):
    instances = odm.Integer(description="Number of retrohunt workers")
    request_time = odm.Optional(odm.Float(description="Time to load metrics"))
    pending_files = odm.integer(description="Files not yet available for searching")
    ingested_last_minute = odm.integer(description="Files ingested in last minute")
    worker_storage_available = odm.integer(description="Free storage for most depleted worker")
    total_storage_available = odm.integer(description="Free storage across workers")
    active_searches = odm.integer(description="Number of currently running searches")
    last_minute_cpu = odm.Float(description="Last minute cpu load across all workers")
    total_memory_used = odm.Float(description="Estimated current memory use across all workers")


@odm.model(description="Model of retrohunt heartbeat message")
class RetrohuntMessage(odm.Model):
    msg = odm.Compound(Heartbeat, description="Heartbeat message for retrohunt")
    msg_loader = odm.Enum(values={LOADER_CLASS}, default=LOADER_CLASS, description="Loader class for message")
    msg_type = odm.Enum(values=MSG_TYPES, default="RetrohuntHeartbeat", description="Type of message")
    sender = odm.Keyword(description="Sender of message")



from assemblyline import odm


INDEX_CATAGORIES = [
    'hot',
    'archive',
    'hot_and_archive',
]


@odm.model(index=True, store=True, description="A search run on stored files.")
class Retrohunt(odm.Model):
    # Metadata
    indices = odm.Enum(INDEX_CATAGORIES, default='hot_and_archive',
                       description="Defines the indices used for this retrohunt job")
    classification = odm.Classification(description="Classification for this retrohunt job")
    search_classification = odm.ClassificationString(description="Maximum classification of results in the search")
    creator = odm.keyword(copyto="__text__", description="User who created this retrohunt job")
    description = odm.Text(copyto="__text__", description="Human readable description of this retrohunt job")
    expiry_ts = odm.Optional(odm.Date(store=False), description="Expiry timestamp of this retrohunt job")

    start_group = odm.long(description="Earliest expiry group this search will include")
    end_group = odm.long(description="Latest expiry group this search will include")
    
    created_time = odm.date(description="Start time for the search.")
    started_time = odm.date(description="Start time for the search.")
    completed_time = odm.Optional(odm.Date(store=False), description="Time that the search ended")
    
    # Search data
    key = odm.keyword(description="Unique code identifying this retrohunt job")
    raw_query = odm.keyword(store=False, description="Text of filter query derived from yara signature")
    yara_signature = odm.keyword(copyto="__text__", store=False, description="Text of original yara signature run")

    # Completion data
    errors = odm.sequence(odm.keyword(store=False), store=False,
                          description="List of error messages that occured during the search")
    warnings = odm.sequence(odm.keyword(store=False), store=False,
                            description="List of warning messages that occured during the search")
    finished = odm.boolean(default=False, description="Boolean that indicates if this retrohunt job is finished")
    truncated = odm.boolean(default=False, description="Indicates if the list of hits been truncated at some limit")


@odm.model(index=True, store=True, description="A hit encountered during a retrohunt search.")
class RetrohuntHit(odm.Model):
    key = odm.keyword(description="Unique code indentifying this hit")
    classification = odm.Classification(description="Classification string for the retrohunt job and results list")
    sha256 = odm.SHA256()
    expiry_ts = odm.Optional(odm.Date(store=False), description="Expiry for this entry.")
    search = odm.keyword()



from assemblyline import odm

MSG_TYPES = {"ScalerStatusHeartbeat"}
LOADER_CLASS = "assemblyline.odm.messages.scaler_status_heartbeat.ScalerStatusMessage"


@odm.model(description="Service Status Model")
class Status(odm.Model):
    running = odm.Integer(description="Number of instances running")
    target = odm.Integer(description="Target scaling for service")
    minimum = odm.Integer(description="Minimum number of instances")
    maximum = odm.Integer(description="Maximum number of instances")
    dynamic_maximum = odm.Integer(description="Dynamic maximum number of instances")
    queue = odm.Integer(description="Service queue")
    pressure = odm.Float(description="Service pressure")
    duty_cycle = odm.Float(description="Duty Cycle")


@odm.model(description="Hearbeat Model")
class Heartbeat(odm.Model):
    service_name = odm.Keyword(description="Name of service")
    metrics = odm.Compound(Status, description="Status of service")


@odm.model(description="Model of Scaler's Status Heartbeat Message")
class ScalerStatusMessage(odm.Model):
    msg = odm.Compound(Heartbeat, description="Heartbeat message")
    msg_loader = odm.Enum(values={LOADER_CLASS}, default=LOADER_CLASS, description="Loader class for message")
    msg_type = odm.Enum(values=MSG_TYPES, default="ScalerStatusHeartbeat", description="Type of message")
    sender = odm.Keyword(description="Sender of message")



from redis import Redis

from assemblyline.remote.datatypes.hash import Hash
from assemblyline.remote.datatypes.queues.named import NamedQueue


DISPATCH_TASK_ASSIGNMENT = 'dispatcher-tasks-assigned-to-'
TASK_ASSIGNMENT_PATTERN = DISPATCH_TASK_ASSIGNMENT + '*'
DISPATCH_START_EVENTS = 'dispatcher-start-events-'
DISPATCH_RESULT_QUEUE = 'dispatcher-results-'
DISPATCH_COMMAND_QUEUE = 'dispatcher-commands-'
DISPATCH_DIRECTORY = 'dispatchers-directory'


class Dispatcher:
    """A utility class for fetching information about the dispatchers running in the system."""
    @staticmethod
    def all_instances(persistent_redis: Redis) -> list[str]:
        """List all dispatchers who have created a listing for themselves."""
        return Hash(DISPATCH_DIRECTORY, host=persistent_redis).keys()

    @staticmethod
    def instance_assignment_size(persistent_redis: Redis, instance_id: str):
        """Get the number of submissions assigned to a given dispatcher instance."""
        return Hash(DISPATCH_TASK_ASSIGNMENT + instance_id, host=persistent_redis).length()

    @staticmethod
    def instance_assignment(persistent_redis: Redis, instance_id: str) -> list[str]:
        """List the submissions assigned to a given dispatcher instance."""
        return Hash(DISPATCH_TASK_ASSIGNMENT + instance_id, host=persistent_redis).keys()

    @staticmethod
    def all_queue_lengths(redis: Redis, instance_id: str):
        """Get the queue lengths for a given dispatcher instance."""
        return {
            'start': NamedQueue(DISPATCH_START_EVENTS + instance_id, host=redis).length(),
            'result': NamedQueue(DISPATCH_RESULT_QUEUE + instance_id, host=redis).length(),
            'command': NamedQueue(DISPATCH_COMMAND_QUEUE + instance_id, host=redis).length()
        }


import re
from copy import copy
from enum import Enum
from typing import Literal, Union, overload

import chardet

# Reference: https://unicode.org/reports/tr9/#Directional_Formatting_Characters
class DirectionalFormattingCharacter(Enum):
    LRE = u'\u202A' # Left-to-Right Embedding
    RLE = u'\u202B' # Right-to-Left Embedding
    PDF = u'\u202C' # Pop Directional Formatting
    LRO = u'\u202D' # Left-to-Right Override
    RLO = u'\u202E' # Right-to-Left Override
    LRI = u'\u2066' # Left-to-Right Isolate
    RLI = u'\u2067' # Right-to-Left Isolate
    FSI = u'\u2068' # First Strong Isolate
    PDI = u'\u2069' # Pop Directional Isolate
    LRM = u'\u200E' # Left-to-Right Mark
    RLM = u'\u200F' # Right-to-Left Mark
    ALM = u'\u061C' # Arabic Letter Mark

CONTROL_CHARS = []
EO_CONTROL_CHARS = []
I_CONTROL_CHARS = []

for c in DirectionalFormattingCharacter:
    CONTROL_CHARS.append(c.value)
    if c in [DirectionalFormattingCharacter.LRE, DirectionalFormattingCharacter.RLE,
             DirectionalFormattingCharacter.LRO, DirectionalFormattingCharacter.RLO]:
        EO_CONTROL_CHARS.append(c.value)
    elif c in [DirectionalFormattingCharacter.LRI, DirectionalFormattingCharacter.RLI,
               DirectionalFormattingCharacter.FSI]:
        I_CONTROL_CHARS.append(c.value)

def remove_bidir_unicode_controls(in_str: str):
    # noinspection PyBroadException
    try:
        no_controls_str = ''.join(c for c in in_str if c not in CONTROL_CHARS)
    except Exception:
        no_controls_str = in_str

    return no_controls_str


def wrap_bidir_unicode_string(uni_str: Union[str, bytes]) -> Union[str, bytes]:
    """
    Wraps str in a LRE (Left-to-Right Embed) unicode control
    Guarantees that str can be concatenated to other strings without
    affecting their left-to-right direction
    """

    if len(uni_str) == 0 or isinstance(uni_str, bytes):  # Not str, return it unchanged
        return uni_str

    re_obj = re.search(rf"[{''.join(CONTROL_CHARS)}]", uni_str)
    if re_obj is None or len(re_obj.group()) == 0:  # No unicode bidir controls found, return string unchanged
        return uni_str

    # Parse str for unclosed bidir blocks
    idf_count = 0   # Isolate Directional Formatting Count
    eodf_count = 0  # Embedding and Override Directional Formatting Count

    for letter in uni_str:
        # Look for block open with embedded or override characters
        if letter in EO_CONTROL_CHARS:
            eodf_count += 1
        # Look for block close with embedded or override characters
        elif letter == DirectionalFormattingCharacter.PDF.value:
            if eodf_count > 0:
                eodf_count -= 1
        # Look for block open with isolate characters
        elif letter in I_CONTROL_CHARS:
            idf_count += 1
        # Look for block close with isolate characters
        elif letter == DirectionalFormattingCharacter.PDI.value:
            if idf_count > 0:
                idf_count -= 1

    # close all bidir blocks
    if eodf_count > 0:
        uni_str += (DirectionalFormattingCharacter.PDF.value * eodf_count)
    if idf_count > 0:
        uni_str += (DirectionalFormattingCharacter.PDI.value * idf_count)

    # Final wrapper (LTR block) to neutralize any Marks (u+200E, u+200F and u+061C)
    uni_str = DirectionalFormattingCharacter.LRE.value + uni_str + DirectionalFormattingCharacter.PDF.value

    return uni_str


# According to wikipedia, RFC 3629 restricted UTF-8 to end at U+10FFFF.
# This removed the 6, 5 and (irritatingly) half of the 4 byte sequences.
#
# The start byte for 2-byte sequences should be a value between 0xc0 and
# 0xdf but the values 0xc0 and 0xc1 are invalid as they could only be
# the result of an overlong encoding of basic ASCII characters. There
# are similar restrictions on the valid values for 3 and 4-byte sequences.
_valid_utf8 = re.compile(rb"""((?:
    [\x09\x0a\x0d\x20-\x7e]|         # 1-byte (ASCII excluding control chars).
    [\xc2-\xdf][\x80-\xbf]|          # 2-bytes (excluding overlong sequences).
    [\xe0][\xa0-\xbf][\x80-\xbf]|    # 3-bytes (excluding overlong sequences).

    [\xe1-\xec][\x80-\xbf]{2}|       # 3-bytes.
    [\xed][\x80-\x9f][\x80-\xbf]|    # 3-bytes (up to invalid code points).
    [\xee-\xef][\x80-\xbf]{2}|       # 3-bytes (after invalid code points).

    [\xf0][\x90-\xbf][\x80-\xbf]{2}| # 4-bytes (excluding overlong sequences).
    [\xf1-\xf3][\x80-\xbf]{3}|       # 4-bytes.
    [\xf4][\x80-\x8f][\x80-\xbf]{2}  # 4-bytes (up to U+10FFFF).
    )+)""", re.VERBOSE)


def _escape(t, reversible=True):
    if t[0] % 2:
        return t[1].replace(b'\\', b'\\\\') if reversible else t[1]
    else:
        return b''.join((b'\\x%02x' % x) for x in t[1])


def dotdump(s):
    if isinstance(s, str):
        s = s.encode()
    return ''.join(['.' if x < 32 or x > 126 else chr(x) for x in s])


def escape_str(s, reversible=True, force_str=False) -> str:
    if isinstance(s, bytes):
        return escape_str_strict(s, reversible)
    elif not isinstance(s, str):
        if force_str:
            return str(s)
        return s

    try:
        return escape_str_strict(s.encode('utf-16', 'surrogatepass').decode('utf-16').encode('utf-8'), reversible)
    except Exception:
        return escape_str_strict(s.encode('utf-8', errors="backslashreplace"), reversible)


# Returns a string (str) with only valid UTF-8 byte sequences.
def escape_str_strict(s: bytes, reversible=True) -> str:
    escaped = b''.join([_escape(t, reversible)
                        for t in enumerate(_valid_utf8.split(s))])
    return escaped.decode('utf-8')


@overload
def safe_str(s: object, force_str: Literal[True]) -> str: ...


@overload
def safe_str(s: Union[str, bytes], force_str: Literal[False] = False) -> str: ...


def safe_str(s, force_str=False):
    return escape_str(s, reversible=False, force_str=force_str)


def is_safe_str(s) -> bool:
    return escape_str(copy(s), reversible=False) == s


# noinspection PyBroadException
def translate_str(s: Union[str, bytes], min_confidence=0.7) -> dict:
    if not isinstance(s, (str, bytes)):
        raise TypeError(f'Expected str or bytes got {type(s)}')

    if isinstance(s, str):
        s = s.encode("utf-8")

    try:
        r = chardet.detect(s)
    except Exception:
        r = {'confidence': 0.0, 'encoding': None, 'language': None}

    if r['confidence'] > 0 and r['confidence'] >= min_confidence:
        try:
            t: Union[bytes, str] = s.decode(r['encoding'])
        except Exception:
            t = s
    else:
        t = s

    r['converted'] = safe_str(t)
    r['encoding'] = r['encoding'] or 'unknown'
    r['language'] = r['language'] or 'unknown'

    return r


# This method not really necessary. More to stop people from rolling their own.
def unescape_str(s):
    return s.decode('string_escape')


def truncate(data: Union[bytes, str], length: int = 100) -> str:
    """
    This method is a helper used to avoid cluttering output
    :param data: The buffer that will be determined if it needs to be sliced
    :param length: The limit of characters to the buffer
    :return str: The potentially truncated buffer
    """
    # We need to force the value to be a string because the output is a string
    string = safe_str(data, force_str=True)
    if len(string) > length:
        return string[:length] + '...'
    return string


class NamedConstants(object):

    def __init__(self, name, string_value_list):
        self._name = name
        self._value_map = dict(string_value_list)
        self._reverse_map = dict([(s[1], s[0]) for s in string_value_list])

        # we also import the list as attributes so things like
        # tab completion and introspection still work.
        for s, v in self._value_map.items():
            setattr(self, s, v)

    def name_for_value(self, v):
        return self._reverse_map[v]

    def contains_value(self, v):
        return v in self._reverse_map

    def __getitem__(self, s):
        return self._value_map[s]

    def __getattr__(self, s):
        # We implement our own getattr mainly to provide the better exception.
        return self._value_map[s]


class StringTable(object):

    def __init__(self, name, string_value_list):
        self._name = name
        self._value_map = dict(string_value_list)
        self._reverse_map = dict([(s[1], s[0]) for s in string_value_list])

        # we also import the list as attributes so things like
        # tab completion and introspection still work.
        for s in self._value_map.keys():
            setattr(self, s, s)

    def name_for_value(self, v):
        return self._reverse_map[v]

    def contains_string(self, s):
        return s in self._reverse_map

    def contains_value(self, v):
        return v in self._value_map

    def __getitem__(self, s):
        if s in self._value_map:
            return s
        raise AttributeError("Invalid value for %s (%s)" % (self._name, s))

    def __getattr__(self, s):
        # We implement our own getattr mainly to provide the better exception.
        if s in self._value_map:
            return s
        raise AttributeError("Invalid value for %s (%s)" % (self._name, s))

    def keys(self) -> list:
        return list(self._value_map.keys())



# cython: language_level=3

# noinspection PyUnresolvedReferences
from libc.string cimport memset

def counts(b, c, d=None):
    if d is None:
        d = {}
    cdef long long t[256]
    cdef unsigned char* s = b
    cdef int l = c
    cdef int i = 0

    memset(t, 0, 256 * sizeof(long long))

    for k, v in d.iteritems():
        t[k] = v

    while i < l:
        t[s[i]] += 1
        i += 1

    return {i: t[i] for i in range(256) if t[i]}

def counts_old(s, d=None):
    if d is None:
        d = {}
    cdef int i
    cdef int t[256]

    memset(t, 0, 256 * sizeof(int))

    for k, v in d.iteritems():
        t[k] = v

    for c in s:
        t[ord(c)] += 1

    return {i: t[i] for i in range(256) if t[i]}



from assemblyline import odm
from assemblyline.common import forge
from assemblyline.common.str_utils import StringTable
from assemblyline.odm.models.user import ACL_MAP, USER_ROLES


APIKEY_ID_DELIMETER = "+"
APIKEY_ID_FORMAT = "{}"+ APIKEY_ID_DELIMETER + "{}"
FORBIDDEN_APIKEY_CHARACTERS = '[+@!#$%^&*()<>?/\|}{~:]'



@odm.model(index=True, store=True, description="Model of Apikey")
class Apikey(odm.Model):
    acl = odm.List(odm.Enum(values=ACL_MAP.keys()), description="Access Control List for the API key")
    password = odm.Keyword(description="BCrypt hash of the password for the apikey")
    roles = odm.List(odm.Enum(values=USER_ROLES), default=[], description="List of roles tied to the API key")
    uname = odm.Keyword(copyto="__text__", description="Username")
    key_name = odm.Keyword(copyto="__text__", description="Name of the key")
    creation_date = odm.Date(default="NOW", description="The date this API key is created.")
    expiry_ts = odm.Optional(odm.Date(), description="Expiry timestamp.")
    last_used =odm.Optional(odm.Date(), description="The last time this API key was used.")

def get_apikey_id(keyname:str , uname:str):
    return APIKEY_ID_FORMAT.format(keyname, uname)

def split_apikey_id(key_id: str):
    data = key_id.split(APIKEY_ID_DELIMETER)
    username = data[1]
    keyname = data[0]

    return keyname, username



from assemblyline import odm
from assemblyline.odm.models.ontology.file import File
from assemblyline.odm.models.ontology.results.process import Process, ObjectID
from assemblyline.common.dict_utils import get_dict_fingerprint_hash

OID_PARTS = ['source_ip', 'source_port',
             'destination_ip', 'destination_port',
             'transport_layer_protocol', 'connection_type']


REQUEST_METHODS = [
    # Standard HTTP methods
    "GET", "POST", "PUT", "DELETE", "HEAD", "CONNECT", "OPTIONS", "TRACE", "PATCH",
    # WebDAV HTTP methods
    "BCOPY", "BDELETE", "BMOVE", "BPROPFIND", "BPROPPATCH", "COPY", "DELETE", "LOCK", "MKCOL", "MOVE",
    "NOTIFY", "POLL", "PROPFIND", "PROPPATCH", "SEARCH", "SUBSCRIBE", "UNLOCK", "UNSUBSCRIBE", "X-MS-ENUMATTS"
]

# https://en.wikipedia.org/wiki/List_of_DNS_record_types
LOOKUP_TYPES = ["A", "AAAA", "AFSDB", "APL", "CAA", "CDNSKEY", "CDS", "CERT", "CNAME", "CSYNC", "DHCID", "DLV",
                "DNAME", "DNSKEY", "DS", "EUI48", "EUI64", "HINFO", "HIP", "HTTPS", "IPSECKEY", "KEY", "KX", "LOC",
                "MX", "NAPTR", "NS", "NSEC", "NSEC3", "NSEC3PARAM", "OPENPGPKEY", "PTR", "RRSIG", "RP", "SIG",
                "SMIMEA", "SOA", "SRV", "SSHFP", "SVCB", "TA", "TKEY", "TLSA", "TSIG", "TXT", "URI", "ZONEMD"]


@odm.model(description="Details for a DNS request")
class NetworkDNS(odm.Model):
    domain = odm.Domain(description="The domain requested")
    resolved_ips = odm.Optional(odm.List(odm.IP()), description="A list of IPs that were resolved")
    resolved_domains = odm.Optional(odm.List(odm.Domain()), description="A list of domains that were resolved")
    lookup_type = odm.Enum(values=LOOKUP_TYPES, description="The type of DNS request")


@odm.model(description="Details for an HTTP request")
class NetworkHTTP(odm.Model):
    request_uri = odm.URI(description="The URI requested")
    request_headers = odm.Mapping(odm.Json(), description="Headers included in the request")
    request_method = odm.Enum(values=REQUEST_METHODS, description="The method of the request")
    response_headers = odm.Mapping(odm.Json(), description="Headers included in the response")
    request_body = odm.Optional(odm.Text(), description="The body of the request")
    response_status_code = odm.Optional(odm.Integer(), description="The status code of the response")
    response_body = odm.Optional(odm.Text(), description="The body of the response")
    response_content_fileinfo = odm.Optional(
        odm.Compound(File),
        description="The file information of the response content")
    response_content_mimetype = odm.Optional(
        odm.Text(), description="The response content mimetype returned by the server")


@odm.model(description="Details for an SMTP request")
class NetworkSMTP(odm.Model):
    mail_from = odm.Email(description="Sender of the email")
    mail_to = odm.List(odm.Email(), description="Recipients of the email")
    attachments = odm.Optional(odm.List(odm.Compound(File)), description="The file information about the attachments")

@odm.model(description="Details for a low-level network connection by IP")
class NetworkConnection(odm.Model):
    objectid = odm.Compound(ObjectID, description="The object ID of the network object")
    destination_ip = odm.Optional(odm.IP(), description="The destination IP of the connection")
    destination_port = odm.Optional(odm.Integer(), description="The destination port of the connection")
    transport_layer_protocol = odm.Optional(odm.Enum(values=["tcp", "udp"]),
                                            description="The transport layer protocol of the connection")
    direction = odm.Optional(odm.Enum(values=["outbound", "inbound", "unknown"]),
                             description="The direction of the network connection")
    process = odm.Optional(odm.Compound(Process), description="The process that spawned the network connection")
    source_ip = odm.Optional(odm.IP(), description="The source IP of the connection")
    source_port = odm.Optional(odm.Integer(), description="The source port of the connection")
    http_details = odm.Optional(odm.Compound(NetworkHTTP), description="HTTP-specific details of request")
    dns_details = odm.Optional(odm.Compound(NetworkDNS), description="DNS-specific details of request")
    smtp_details = odm.Optional(odm.Compound(NetworkSMTP), description="SMTP-specific details of request")
    connection_type = odm.Optional(odm.Enum(values=['http', 'dns', 'tls', 'smtp'],
                                            description="Type of connection being made"))

    def get_oid(data: dict):
        connection_type = data.get('connection_type')
        hash_dict = {key: data.get(key) for key in OID_PARTS}
        oid_prefix = "network"
        if connection_type == "http":
            oid_prefix = "network_http"
            http_details = data.get('http_details', {})

            # Include any details involved in the request for hashing
            hash_dict['http_details'] = {
                field: http_details.get(field)
                for field in NetworkHTTP.fields().keys() if field.startswith('request_')
            }
        elif connection_type == "dns":
            # Include the requested domain as part of the hash
            oid_prefix = "network_dns"
            hash_dict['dns_details'] = {
                'domain': data.get('dns_details', {}).get('domain', None),
                'lookup_type': data.get('dns_details', {}).get('lookup_type', None)
            }
        elif connection_type == "smtp":
            # Include the mail_to, mail_from, and attachement details as part of the hash
            oid_prefix = "network_smtp"
            hash_dict['smtp_details'] = data.get('smtp_details', {})

        return f"{oid_prefix}_{get_dict_fingerprint_hash(hash_dict)}"

    def get_tag(data: dict):
        return f"{data.get('destination_ip')}:{data.get('destination_port')}"



from assemblyline import odm

MSG_TYPES = {"VacuumHeartbeat"}
LOADER_CLASS = "assemblyline.odm.messages.vacuum_heartbeat.VacuumMessage"


@odm.model(description="Vacuum Stats")
class Metrics(odm.Model):
    ingested = odm.Integer(description="Files ingested")
    # protocol = odm.Mapping(odm.Integer())
    safelist = odm.Integer(description="Files safelisted")
    errors = odm.Integer()
    skipped = odm.Integer()


@odm.model(description="Heartbeat Model")
class Heartbeat(odm.Model):
    # instances = odm.Integer(description="Number of instances")
    metrics = odm.Compound(Metrics, description="Vacuum metrics")
    # queues = odm.Compound(Metrics, description="Vacuum queues")


@odm.model(description="Model of Vacuum Heartbeat Message")
class VacuumMessage(odm.Model):
    msg = odm.Compound(Heartbeat, description="Hearbeat message")
    msg_loader = odm.Enum(values={LOADER_CLASS}, default=LOADER_CLASS, description="Loader class for message")
    msg_type = odm.Enum(values=MSG_TYPES, default="VacuumHeartbeat", description="Type of message")
    sender = odm.Keyword(description="Sender of message")



from assemblyline import odm
from assemblyline.common import constants, forge
from assemblyline.odm.models.config import ServiceSafelist
from assemblyline.odm.models.file import URIInfo

Classification = forge.get_classification()

MSG_TYPES = {"Task"}
LOADER_CLASS = "assemblyline.odm.messages.task.TaskMessage"


@odm.model(description="File Information")
class FileInfo(odm.Model):
    magic = odm.Keyword(description="The output from libmagic which was used to determine the tag")
    md5 = odm.MD5(description="MD5 of the file")
    mime = odm.Optional(odm.Keyword(), description="The libmagic mime type")
    sha1 = odm.SHA1(description="SHA1 hash of the file")
    sha256 = odm.SHA256(description="SHA256 hash of the file")
    size = odm.Integer(description="Size of the file in bytes")
    ssdeep = odm.Optional(odm.SSDeepHash(description="SSDEEP hash of the file"))
    tlsh = odm.Optional(odm.Keyword(description="TLSH hash of the file"))
    type = odm.Keyword(description="Type of file as identified by Assemblyline")
    uri_info = odm.Optional(odm.Compound(URIInfo), description="URI structure to speed up specialty file searching")


@odm.model(description="Tag Item")
class TagItem(odm.Model):
    type = odm.Keyword(description="Type of tag item")
    value = odm.Keyword(description="Value of tag item")
    short_type = odm.Keyword(description="Short version of tag type")
    score = odm.Optional(odm.Integer(), description="Score of tag item")


@odm.model(description="Data Item")
class DataItem(odm.Model):
    name = odm.Keyword()
    value = odm.Any()


@odm.model(description="Service Task Model")
class Task(odm.Model):
    sid = odm.UUID(description="Submission ID")
    metadata: dict[str, str] = odm.FlatMapping(odm.MetadataValue(), default={}, description="Metadata associated to the submission")
    min_classification = odm.Classification(description="Minimum classification of the file being scanned")
    fileinfo: FileInfo = odm.Compound(FileInfo, description="File info block")
    filename = odm.Keyword(description="File name")
    service_name = odm.Keyword(description="Service name")
    service_config = odm.Mapping(odm.Any(), default={}, description="Service specific parameters")
    depth = odm.Integer(default=0, description="File depth relative to initital submitted file")
    max_files = odm.Integer(description="Maximum number of files that submission can have")
    ttl = odm.Integer(default=0, description="Task TTL")

    tags = odm.List(odm.Compound(TagItem), default=[], description="List of tags")
    temporary_submission_data = odm.List(odm.Compound(DataItem), default=[], description="Temporary submission data")

    deep_scan = odm.Boolean(default=False, description="Perform deep scanning")

    ignore_cache = odm.Boolean(
        default=False, description="Whether the service cache should be ignored during the processing of this task")

    ignore_recursion_prevention = odm.Boolean(
        default=False,
        description="Whether the service should ignore recursion prevention or not")

    ignore_filtering = odm.Boolean(default=False, description="Should the service filter it's output?")

    priority = odm.Integer(default=1, description="Priority for processing order",
                           min=constants.DROP_PRIORITY, max=constants.MAX_PRIORITY)
    safelist_config = odm.Compound(ServiceSafelist,
                                   description="Safelisting configuration (as defined in global configuration)",
                                   default={'enabled': False})

    @staticmethod
    def make_key(sid, service_name, sha):
        return f"{sid}_{service_name}_{sha}"

    def key(self):
        return Task.make_key(self.sid, self.service_name, self.fileinfo.sha256)


@odm.model(description="Model for Service Task Message")
class TaskMessage(odm.Model):
    msg = odm.Compound(Task, description="Body of the message")
    msg_loader = odm.Enum(values={LOADER_CLASS}, default=LOADER_CLASS,
                          description="Class to use to load the message as an object")
    msg_type = odm.Enum(values=MSG_TYPES, default="Task", description="Type of message")
    sender = odm.Keyword(description="Sender of the message")



from assemblyline import odm

MSG_TYPES = {"MetricsCounter"}
LOADER_CLASS = "assemblyline.odm.messages.metrics.MetricsMessage"


@odm.model(description="Metrics Model")
class Metrics(odm.Model):
    host = odm.Keyword(description="Host that generated metric")
    type = odm.Keyword(description="Type of metric")
    name = odm.Keyword(description="Metric name")
    metrics = odm.Mapping(odm.Integer(), description="Metric value")


@odm.model(description="Model of Metric Message")
class MetricsMessage(odm.Model):
    msg = odm.Compound(Metrics, description="Metrics message")
    msg_loader = odm.Enum(values={LOADER_CLASS}, default=LOADER_CLASS, description="Loader class for message")
    msg_type = odm.Enum(values=MSG_TYPES, default="MetricsCounter", description="Type of message")
    sender = odm.Keyword(description="Sender of message")



from assemblyline import odm
from assemblyline.odm.models.ontology.file import File


@odm.model(index=False, store=False, description="")
class HTTPRedirect(odm.Model):
    from_url = odm.Keyword(description="")
    to_url = odm.Keyword(description="")


@odm.model(index=False, store=False, description="HTTP Task")
class HTTP(odm.Model):
    response_code = odm.Integer(description="The status code of the main page")
    redirection_url = odm.Optional(odm.Keyword(), description="The final page of the requested url")
    redirects = odm.Optional(odm.List(odm.Compound(HTTPRedirect)), description="List of Redirects")
    favicon = odm.Optional(odm.Compound(File), description="The file information of the main favicon")
    title = odm.Optional(odm.Keyword(), description="The title of the main page after any redirection")
